import typing

from fastapi import Request, Response
from fastapi.responses import JSONResponse
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint

from ..backend.fastapi import FastAPIKeycloakAuthBackend


class BaseFastAPIKCAuthentication(BaseHTTPMiddleware):
    backend = FastAPIKeycloakAuthBackend

    def __init__(
        self,
        host: typing.Optional[str] = None,
        realm: typing.Optional[str] = None,
        algorithms: typing.Optional[list[str]] = None,
        audience: typing.Optional[str] = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.host = host or self.kc_host
        self.realm = realm or self.kc_realm
        self.algorithms = algorithms or self.kc_algorithms
        self.audience = audience or self.kc_audience

    def authenticate(self, request: Request) -> typing.Optional[dict]:
        backend = self.backend(
            request=request,
            host=self.host,
            realm=self.realm,
            algorithms=self.algorithms,
            audience=self.audience,
        )
        return backend.authenticate()

    def post_process_claims(
        self,
        claims: typing.Optional[dict],
        request: Request,
    ) -> Request:
        return request

    async def dispatch(
        self,
        request: Request,
        call_next: RequestResponseEndpoint,
    ) -> Response:
        try:
            claims = self.authenticate(request=request)
        except self.backend.AuthError as e:
            return JSONResponse(status_code=401, content={"detail": str(e)})
        request = self.post_process_claims(claims=claims, request=request)
        response = await call_next(request)
        return response



import jwt

from .errors import JWTDecodeError


def verify_token(
    access_token: str,
    public_key: str,
    algorithms: list[str],
    audience: str,
) -> dict:
    """
    Verifies a token with Keycloak and returns the validated claims
    """
    try:
        return jwt.decode(
            access_token,
            key=public_key,
            algorithms=algorithms,
            audience=audience,
        )
    except jwt.InvalidTokenError as e:
        raise JWTDecodeError(str(e))



import typing

from django.contrib.auth import get_user_model
from rest_framework.authentication import BaseAuthentication as DRFBaseAuth
from rest_framework.exceptions import AuthenticationFailed
from rest_framework.request import Request

from ..backend.rest_framework import DRFKeycloakAuthBackend

User = get_user_model()


class BaseDRFKCAuthentication(DRFBaseAuth):
    kc_host: str
    kc_realm: str
    kc_algorithms: list[str]
    kc_audience: str

    keyword = "Bearer"
    backend = DRFKeycloakAuthBackend
    AuthFailedError: Exception = AuthenticationFailed

    def get_or_create_user(self, claims: dict) -> User:
        raise NotImplementedError

    def authenticate(
        self,
        request: Request,
    ) -> typing.Optional[typing.Tuple[User, None]]:
        try:
            backend = self.backend(
                request,
                host=self.kc_host,
                realm=self.kc_realm,
                algorithms=self.kc_algorithms,
                audience=self.kc_audience,
            )
            claims = backend.authenticate()
        except self.backend.AuthError as e:
            raise self.AuthFailedError(e.msg)

        if not claims:
            return None

        user = self.get_or_create_user(claims=claims)
        return user, None

    def authenticate_header(self, request) -> str:
        return self.keyword



import typing
from functools import lru_cache

import requests

from ..errors import PublicKeyNotFound


class BasePublicKeyManager:
    ttl: int = 60 * 60 * 24  # 1 day
    host: str
    realm: str

    def __init__(
        self,
        host: typing.Optional[str] = None,
        realm: typing.Optional[str] = None,
    ):
        self.host = host or ""
        self.realm = realm or ""

    @property
    def url_realm(self) -> str:
        return f"{self.host}/auth/realms/{self.realm}"

    def get_fresh_key_from_upstream(self) -> str:
        response = requests.get(self.url_realm)
        if response.status_code != 200:
            raise PublicKeyNotFound("Keycloak public key not found")
        return response.json().get("public_key")

    def get_fresh_pem_key(self) -> str:
        key = self.get_fresh_key_from_upstream()
        return f"-----BEGIN PUBLIC KEY-----\n{key}\n-----END PUBLIC KEY-----"

    def clear_cache(self, *args, **kwargs):
        self.get_key_from_cache.cache_clear()

    @lru_cache
    def get_key_from_cache(self, *args, **kwargs) -> typing.Optional[str]:
        return self.get_fresh_pem_key()

    def set_key(self, key: str, *args, **kwargs) -> str:
        ...

    def get_or_set_key(self, *args, **kwargs) -> str:
        key = self.get_key_from_cache(*args, **kwargs)
        if key:
            return key
        key = self.get_fresh_pem_key()
        self.set_key(key, *args, **kwargs)
        return key

    def get_key(self, force: bool = False, *args, **kwargs) -> str:
        if force:
            self.clear_cache(*args, **kwargs)
        return self.get_or_set_key(*args, **kwargs)



import typing
from functools import lru_cache

from ..manager.base import BasePublicKeyManager
from ..utils import verify_token


class BaseTokenVerifier:
    manager: typing.Type[BasePublicKeyManager]

    def __init__(
        self,
        access_token: str,
        host: str,
        realm: str,
        algorithms: list[str],
        audience: str,
    ):
        self.access_token = access_token
        self.host = host
        self.realm = realm
        self.algorithms = algorithms
        self.audience = audience

    @lru_cache
    def get_claims(self, force=False, *args, **kwargs) -> dict:
        """
        Verify the token using Public Key
        """
        manager = self.manager(host=self.host, realm=self.realm)
        public_key = manager.get_key(force=force)
        return verify_token(
            access_token=self.access_token,
            public_key=public_key,
            algorithms=self.algorithms,
            audience=self.audience,
        )



class KeyCloakBaseError(Exception):
    """
    Base class for all Keycloak errors
    """

    def __init__(self, msg, *args, **kwargs):
        self.msg = msg

    def __str__(self):
        return str(self.msg)

    def __repr__(self):
        return f"{self.__class__.__name__}({self})"


class KeycloakError(KeyCloakBaseError):
    """
    Base class for all Keycloak errors
    """


class PublicKeyNotFound(KeycloakError):
    """
    Base class for all Keycloak errors
    """


class JWTDecodeError(KeycloakError):
    """
    Error while decoding the JWT token
    """
